//! Work around features missing from stable.

use arrayvec::ArrayVec;

//--------------------------------------------------------------------------------------------------
// std::array::try_from_fn

pub(crate) fn try_from_fn<T, E, const N: usize>(
    mut f: impl FnMut(usize) -> Result<T, E>,
) -> Result<[T; N], E> {
    // Use small stack-allocated vector as temporary initializer
    let mut res = ArrayVec::<T, N>::new();
    for i in 0..N {
        let next = f(i)?;
        res.push(next);
    }
    Ok(res.into_inner().unwrap_or_else(|_| unreachable!())) // (to avoid T: Debug)
}

//--------------------------------------------------------------------------------------------------
// std::array::try_map

pub(crate) trait TryMap<I, const N: usize>: IntoIterator<Item = I> + Sized {
    fn try_map<E, O>(self, f: impl FnMut(I) -> Result<O, E>) -> Result<[O; N], E>;
}

impl<I, const N: usize> TryMap<I, N> for [I; N] {
    fn try_map<E, O>(self, f: impl FnMut(I) -> Result<O, E>) -> Result<[O; N], E> {
        let mut f = f;
        // Same strategy as above.
        let mut res = ArrayVec::<O, N>::new();
        for a in self {
            let next = f(a)?;
            res.push(next);
        }
        Ok(res.into_inner().unwrap_or_else(|_| unreachable!()))
    }
}

//--------------------------------------------------------------------------------------------------
// std::slice::split_once (take this opportunity to simplify for our use case)

pub(crate) trait SplitOnce<T> {
    type Chunk;
    fn split_once(self, pred: T) -> Option<(Self::Chunk, Self::Chunk)>;
}
impl<'s> SplitOnce<u8> for &'s [u8] {
    type Chunk = &'s [u8];
    fn split_once(self, pred: u8) -> Option<(Self::Chunk, Self::Chunk)> {
        let index = self.iter().position(|&c| c == pred)?;
        Some((&self[..index], &self[index + 1..]))
    }
}
