//! Describe writer threads and how to handle them.
//! This thread receives fastq blocks rendered as raw buffers
//! and writers from the dispatcher thread,
//! and is just in charge of pushing the buffers through the writer
//! before sending it back for recycling and redispatching.
//! R1 and R2 writers are paired so that paired files are consistently ordered.

use std::{
    io::Write,
    sync::mpsc::{Receiver, Sender},
    thread::JoinHandle,
};

use snafu::{ResultExt, Snafu};

use crate::{io::Context, Block, ContextualizedWriter, SampleCode};

/// Messages received.
pub(crate) enum Input<W: Write> {
    Write {
        /// Which sample files to output to.
        sample_code: SampleCode,
        /// Formatted [R1, R2] blocks, ready to be zipped.
        buffers: [Vec<u8>; 2],
        /// The corresponding writers to push the buffers within.
        writers: [ContextualizedWriter<W>; 2],
        /// The number of fastq blocks in each buffer.
        n_blocks: usize,
        /// Raise to flush after write, meaning that the writers can be consumed/dropped.
        flush: bool,
    },
    Terminate,
}

/// Messages sent.
pub(crate) enum Output<W: Write> {
    Written {
        /// Thread id, useful for the main thread to mark this one ready for a new job.
        id: usize,
        /// Sample files just written, useful for the main thread to attribute the writers.
        sample_code: SampleCode,
        /// Consumed, cleared buffers, ready for recycling.
        buffers: [Vec<u8>; 2],
        /// These writers finished the job, ready for redispatching.
        writers: [ContextualizedWriter<W>; 2],
    },
    Flushed {
        id: usize,
        buffers: [Vec<u8>; 2],
    },
}

// Owned by the writer thread.
pub(crate) struct Inner<W: Write> {
    pub(crate) id: usize,
    pub(crate) input: Receiver<Input<W>>,
    pub(crate) output: Sender<Output<W>>,
}

// Owned by the main thread.
pub(crate) struct Outer<W: Write> {
    pub(crate) handle: JoinHandle<Result<(), Error>>,
    pub(crate) input: Sender<Input<W>>,
    // Only one common output, not one per thread.
}

impl<W: Write> Inner<W> {
    pub(crate) fn run(self) -> Result<(), Error> {
        use Input as I;
        let Self { id, input, output } = self;
        // Messages are owned bytes slices representing fasta blocks to write.
        for message in &input {
            match message {
                I::Write {
                    sample_code,
                    mut buffers,
                    mut writers,
                    n_blocks,
                    flush,
                } => {
                    for (buffer, writer) in buffers.iter_mut().zip(writers.iter_mut()) {
                        writer
                            .write_all(buffer)
                            .with_context(|_| WriteLineErr { context: &*writer })?;
                        writer.i_line += Block::n_lines() * n_blocks as u64;
                        buffer.clear();
                        if flush {
                            writer
                                .flush()
                                .with_context(|_| FlushErr { context: &*writer })?;
                        }
                    }
                    // Send back for recycling.
                    #[allow(clippy::match_bool)] // (flush last)
                    output
                        .send(match flush {
                            false => Output::Written {
                                id,
                                sample_code,
                                buffers,
                                writers,
                            },
                            true => Output::Flushed { id, buffers },
                        })
                        .map_err(|e| Error::Message { erm: e.to_string() })?;
                }
                I::Terminate => break,
            }
        }
        Ok(())
    }
}

#[derive(Debug, Snafu)]
#[snafu(context(suffix(Err)))]
pub enum Error {
    #[snafu(display("{context}IO error when writing line:{source}"))]
    WriteLine {
        source: std::io::Error,
        context: Context,
    },
    #[snafu(display("Message could not be sent from writer thread to main thread:\n{erm}"))]
    Message {
        erm: String, // Rendered underlying error message.
    },
    #[snafu(display("{context}IO error when flushing to output file:\n{source}"))]
    Flush {
        source: std::io::Error,
        context: Context,
    },
}
